// MeTA~Fract :: 2002-2004 Tito Dal Canton, tito@dalcanton.it
#include <stdlib.h>
#include <stdio.h>
#include <SDL.h>
#include "julia.h"
#include "graph-io.h"

#define PUT_PIXEL(x, y, c)	*(((unsigned char *)main_surface->pixels) + (y) * main_surface->pitch + (x)) = c

extern int		width, height;
extern double	gSupX, gInfX, gSupY, gInfY;
static double	incremx, incremy;
extern int		gMaxIterations;

// riporta il piano complesso ai valori iniziali, in modo da
// avere una visione completa dell'intero insieme
void FullJuliaSet(void)
{
	double	aspect;
	
	aspect = (double)height / (double)width;
    if (width > height) {
		gInfY = -2; gSupY = 2;
		gInfX = gInfY / aspect;
		gSupX = gSupY / aspect;
	} else {
		gInfX = -2; gSupX = 2;
		gInfY = aspect * gInfX;
		gSupY = aspect * gSupX;
	}
}

// disegna l'insieme di Julia
void RenderJulia(double pr, double pi)
{
	int				x, y, max_y = (height >> 1);
	double			r, i;
	unsigned char	*pixptr = (unsigned char *)main_surface->pixels;

	// aggiorna i parametri
	incremx = (gSupX - gInfX) / width;
	incremy = (gSupY - gInfY) / height;
	
//	SDL_LockSurface(main_surface);
	
	// renderizza il frattale
	// (in realta' ne renderizziamo meta', e la parte inferiore
	// la otteniamo per riflessione ;-)
	i = gInfY;
	for (y = 0; y < max_y; y++) {
		r = gInfX;
		for (x = 0; x < width; x++) {
			int		c = 0, color;
			double	r2 = r * r, i2 = i * i;
			double	r_ = r, i_ = i;

			// calcola il colore del pixel tramite iterazioni della
			// formula di Julia Zn = Z(n-1)^2 + C
			while (((r2 + i2) < 4.0) && (c <= gMaxIterations)) {
				i_ = 2 * i_ * r_ + pi;
				r_ = r2 - i2 + pr;
				r2 = r_ * r_;
				i2 = i_ * i_;
				c++;
			}

			if (c < (gMaxIterations - 1))
				color = (c % (kNumOfColors - 2) + 2);
			else
				color = 0;
			*(pixptr + x) = color;
			PUT_PIXEL(width-x, height-y-1, color);
			r += incremx;
		}
		i += incremy;
		pixptr += main_surface->pitch;
	}
	
	SDL_UpdateRect(main_surface, 0, 0, 0, 0);
//	SDL_UnlockSurface(main_surface);
}

