// MeTA~Fract :: 2002-2004 Tito Dal Canton, tito@dalcanton.it
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <math.h>
#include <string.h>
#include <SDL.h>
#include "mandelbrot.h"
#include "julia.h"
#include "graph-io.h"
#include "font-stuff.h"

#define RGB_COLOR(r, g, b)	SDL_MapRGB(offscreen_surface->format, (r), (g), (b))
#define BIT_DEPTH			8
typedef unsigned char		pixel;

enum {
	MANDELBROT_SET,
	JULIA_SET
};

static void main_loop(void);
void FrameRect(int left, int top, int right, int bottom, int color);
static void SaveAs(void);
static void ask_string(char *prompt, char *str, int maxlen, int x, int y);
static int ask_int(char *prompt, int x, int y);
static void parse_args(int argc, char *argv[]);
static void show_help(void);

SDL_Surface		*main_surface;
int				sdl_flags = SDL_HWPALETTE | SDL_HWSURFACE;
int				width = 600, height = 600;
SDL_Color		palette[256];
double			gInfX, gInfY, gSupX, gSupY;
int				gMaxIterations = 1000;
int				whichPalette = 0;
int				set = MANDELBROT_SET;
int				mandel_algo = BOUNDARY_TRACE;

int main(int argc, char *argv[])
{
	char	title[100] = "Insieme di ";

	// esamina i parametri
	parse_args(argc, argv);
	
	// inizializza SDL
	if (SDL_Init(SDL_INIT_VIDEO) < 0) {
		fprintf(stderr, "Non posso inizializzare SDL (%s)\n", SDL_GetError());
		exit(1);
	}
	switch (set) {
		case MANDELBROT_SET:	strcat(title, "Mandelbrot");	break;
		case JULIA_SET:			strcat(title, "Julia");			break;
		default:				strcat(title, "qualche pazzo sconosciuto");	break;
	}
	SDL_WM_SetCaption(title, "metafract-icon.png");
	main_surface = SDL_SetVideoMode(width, height, BIT_DEPTH, sdl_flags);
	
	// cambia la palette
	ChangePalette(whichPalette, main_surface);

	// inizializza il generatore di frattali
	switch (set) {
		case MANDELBROT_SET:
			FullMandelbrotSet(width, height);
			ReallocPixelStatusBuf();
			RenderMandelbrot(main_surface, mandel_algo);
			break;
		case JULIA_SET:
			FullJuliaSet();
			RenderJulia(0, 0);
			break;
		default:
			break;
	}
	
	main_loop();
	
	// fine
	SDL_Quit();
	return 0;
}

static void main_loop(void)
{
	int			the_end = 0;	
	int			tracking_for_zoom = 0, track_origin_x, track_origin_y;
	SDL_Surface	*save_buf = NULL;
	
	// ciclo principale
	while (!the_end) {
		SDL_Event	event;
		int 		new_x, new_y;

		event.type = SDL_NOEVENT;
		SDL_PollEvent(&event);
		switch (event.type) {
			case SDL_MOUSEBUTTONDOWN:
				if (set == MANDELBROT_SET) {
					// mouse cliccato, iniziamo a tracciare il rettangolo di zoom
					SDL_PixelFormat	*fmt;
				
					tracking_for_zoom = 1;
					track_origin_x = event.button.x;
					track_origin_y = event.button.y;
					fmt = main_surface->format;
					save_buf = SDL_CreateRGBSurface(SDL_HWSURFACE, width, height, BIT_DEPTH,
														fmt->Rmask, fmt->Gmask, fmt->Bmask, fmt->Amask);
					if (save_buf == NULL) {
						printf("SDL_CreateRGBSurface fallita (%s)\n", SDL_GetError());
					} else {
						SDL_SetColors(save_buf, palette, 0, 256);
						SDL_BlitSurface(main_surface, NULL, save_buf, NULL);
						SDL_UpdateRect(save_buf, 0, 0, width, height);
					}
				}
				break;
			case SDL_MOUSEMOTION:
				if (tracking_for_zoom && set == MANDELBROT_SET) {
					// questo fa funzionare il rettangolo di zoom finche'
					// il mouse e' cliccato
					new_x = event.motion.x;
					new_y = event.motion.y;
					if (new_x < width && new_y < height) {
						double		win_aspect = (double)height / (double)width;
						int			delta_x, delta_y;
						SDL_Rect	r;

						delta_x = new_x - track_origin_x + 1;
						delta_y = new_y - track_origin_y + 1;
						// aggiusta le dimensioni del rettangolo in modo
						// che siano in proporzione alle dimensioni della
						// finestra del frattale (altrimenti il frattale
						// verrebbe distorto)
						if (delta_x > delta_y) {
							delta_y = delta_x * win_aspect;
						} else {
							delta_x = delta_y / win_aspect;
						}
						new_x = delta_x + track_origin_x - 1;
						new_y = delta_y + track_origin_y - 1;
						r.x = track_origin_x;	r.y = track_origin_y;
						r.w = delta_x;			r.h = delta_y;
						// disegna il rettangolo
						SDL_BlitSurface(save_buf, &r, main_surface, &r);
						FrameRect(track_origin_x, track_origin_y, new_x, new_y, 0);
						SDL_UpdateRect(main_surface, track_origin_x, track_origin_y, delta_x, delta_y);
					}
				}
				break;
			case SDL_MOUSEBUTTONUP:
				if (set == MANDELBROT_SET) {
					// mouse rilasciato, calcola la posizione del rettangolo
					// di zoom nel piano complesso e renderizza il frattale
					// in base alla nuova posizione
					double	inf_x, inf_y, sup_x, sup_y;
					
					tracking_for_zoom = 0;
					SDL_FreeSurface(save_buf);
					save_buf = NULL;
					if ((new_x - track_origin_x) > 3 &&
						(new_y - track_origin_y) > 3) {
						inf_x = track_origin_x * (gSupX - gInfX) / width + gInfX;
						inf_y = track_origin_y * (gSupY - gInfY) / height + gInfY;
						sup_x = new_x * (gSupX - gInfX) / width + gInfX;
						sup_y = new_y * (gSupY - gInfY) / height + gInfY;
						gInfX = inf_x;	gSupX = sup_x;
						gInfY = inf_y;	gSupY = sup_y;
						RenderMandelbrot(main_surface, mandel_algo);
					}
				}
				break;
			case SDL_KEYDOWN:
				switch (event.key.keysym.sym) {
					case SDLK_BACKSPACE:
						if (set == MANDELBROT_SET) {
							FullMandelbrotSet(width, height);
							RenderMandelbrot(main_surface, mandel_algo);
						} else if (set == JULIA_SET)
							FullJuliaSet();
						break;
					case SDLK_ESCAPE:
					case SDLK_q:
						the_end = 1;
						break;
					case SDLK_p:
						whichPalette++;
						if (whichPalette >= PALETTE_COUNT)
							whichPalette = 0;
						ChangePalette(whichPalette, main_surface);
						break;
					case SDLK_s:
						SaveAs();
						SDL_SaveBMP(main_surface, "/tmp/metafract-out.bmp");
						break;
					case SDLK_i:
						{
							char	tempstr[64] = "";

							SDL_EnableUNICODE(1);
							sprintf(tempstr, "Iterazioni massime (attualmente %d): ", gMaxIterations);
							gMaxIterations = ask_int(tempstr, 5, 5);
							SDL_EnableUNICODE(0);
							if (set == MANDELBROT_SET)
								RenderMandelbrot(main_surface, mandel_algo);
						}
						break;
					default:
						break;
				}
				break;
			case SDL_QUIT:
				the_end = 1;
				break;
			default:
				if (set == JULIA_SET) {
					int		mx, my;
					double	x, y;
					
					SDL_GetMouseState(&mx, &my);
					x = mx * (gSupX - gInfX) / width + gInfX;
					y = my * (gSupY - gInfY) / height + gInfY;
					RenderJulia(x, y);
				} else {
					SDL_Delay(10);
				}
		}
	}
}

void ChangePalette(int palno, SDL_Surface *surf)
{
	int	i;
	
	palette[0].r = palette[0].g = palette[0].b = 0;
	palette[1].r = palette[1].g = palette[1].b = 255;
	
	switch (palno) {
		case PALETTE_BLUE_GREEN:    // azzurro-verde
			for (i = 2; i < 256; i++) {
				palette[i].r = 40;
				palette[i].g = sin(i * 3.14 / 256.0) * 200 + 56;
				palette[i].b = (cos(i * 3.14 / 256.0) + 1.0) * 40 + 40;
			}
			break;
		case PALETTE_FLAMING:	// rosso e giallo
			for (i = 2; i < 256; i++) {
				palette[i].r = 255;
				palette[i].g = sin(i * 3.14 / 256.0) * 200 + 56;
				palette[i].b = 40;
			}
			break;
		case PALETTE_GRAYYELLOW:	// bande alternate gialle e sfumate in grigio
			for (i = 2; i < 256; i++) {
				palette[i].r = (i % 6 == 0) ? 250 : sin((i-2) * M_PI / 254) * 255;
				palette[i].g = (i % 6 == 0) ? 250 : sin((i-2) * M_PI / 254) * 255;
				palette[i].b = (i % 6 == 0) ?  60 : sin((i-2) * M_PI / 254) * 255;
			}
			break;
		case PALETTE_CYAN:
			for (i = 2; i < 256; i++) {
				palette[i].r = abs(cos(i * 3.14 / 256.0) * 255);
				palette[i].g = sin(i * 3.14 / 256.0) * 200;
				palette[i].b = sin(i * 3.14 / 256.0) * 256;
			}
			break;						
		case PALETTE_BW_BANDS:  // bande bianche e nere
			for (i = 2; i < 256; i++) {
				palette[i].r = (i & 1) * 255;
				palette[i].g = (i & 1) * 255;
				palette[i].b = (i & 1) * 255;
			}
			break;
		case PALETTE_GRAY_BANDS:    // uber-spettacoloso
			for (i = 2; i < 256; i++) {
				int	foobar = (sin(i * 3.14 / 256.0) * 210.0) + 20;
				
				if (i & 1)
					foobar += 25;
				palette[i].r = foobar;
				palette[i].g = foobar;
				palette[i].b = foobar;
			}
			break;
		case PALETTE_JUST_M_BLACK:  // insieme nero, aree equipotenziali bianche (per esplorazioni)
			for (i = 2; i < 256; i++)
				palette[i].r = palette[i].g = palette[i].b = 255;
			break;
		default:
			break;
	}
	SDL_SetColors(surf, palette, 0, 256);
}

void FrameRect(int left, int top, int right, int bottom, int color)
{
	unsigned char	*pbuf = (unsigned char *)main_surface->pixels;
	int				i;
	
	if (left > right) {
		int	temp = right;

		right = left;
		left = temp;
	}
	if (top > bottom) {
		int	temp = bottom;

		bottom = top;
		top = temp;
	}
	memset(pbuf + left + top * width, color, right - left);
	memset(pbuf + left + bottom * width, color, right - left);
	for (i = top; i <= bottom; i++) {
		*(pbuf + left + i * width) = color;
		*(pbuf + right + i * width) = color;
	}
}

// renderizza il frattale e salvalo come file BMP
static void SaveAs(void)
{
	char		filename[256] = "";
	int			zoom, w, h;
	SDL_Surface	*temp_surface = NULL;

	if (set == JULIA_SET) {
		draw_string("Il salvataggio di immagini Julia non  ancora supportato", 5, 5, 1);
		return;
	}
	
	SDL_EnableUNICODE(1);
	zoom = ask_int("Percentuale di zoom: ", 5, 5);
	w = width * zoom / 100;
	h = height * zoom / 100;
	ask_string("Percorso del file:   ", filename, 255, 5, 5 + FONT_HEIGHT);
	
//	printf("Salvo l'immagine in %s, zoom %d\n", filename, zoom);
	draw_string("Abbi pazienza, renderizzo il frattale...", 5, 5 + 2 * FONT_HEIGHT, 1);
	temp_surface = SDL_CreateRGBSurface(SDL_SWSURFACE, w, h, 8, 0, 0, 0, 0);
	if (temp_surface) {
		ChangePalette(whichPalette, temp_surface);
		if (set == MANDELBROT_SET)
			RenderMandelbrot(temp_surface, PIXEL_BY_PIXEL);
//		else if (set == JULIA_SET)
//			RenderJulia(temp_surface);
		SDL_SaveBMP(temp_surface, filename);
		SDL_FreeSurface(temp_surface);
	}
	SDL_EnableUNICODE(0);
	if (set == MANDELBROT_SET)
		RenderMandelbrot(main_surface, mandel_algo);
}

static void ask_string(char *prompt, char *str, int maxlen, int x, int y)
{
	int			i = 0;
	int			blinker = 0, lastblink = 0;
	SDL_Rect	temprect = {x, y, 0, 0};
	
	temprect.w = strlen(prompt) * FONT_WIDTH;
	temprect.h = FONT_HEIGHT;
	SDL_FillRect(main_surface, &temprect, 0);
	draw_string(prompt, x, y, 1);
	x += FONT_WIDTH * strlen(prompt);
	
	while (1) {
		SDL_Event	event;
		
		event.type = SDL_NOEVENT;
		SDL_PollEvent(&event);
		if (event.type == SDL_KEYDOWN) {
			if (event.key.keysym.sym == SDLK_RETURN || i == maxlen) {
				str[i] = 0;
				return;
			} else {
				if ((event.key.keysym.unicode & 0xFF80) == 0) {		// ASCII standard?
					char	c = event.key.keysym.unicode & 0x7F;	// si
					
					if (c > 31) {
						SDL_Rect	eraserect = {x, y, FONT_WIDTH, FONT_HEIGHT};
						
						str[i] = c;
						SDL_FillRect(main_surface, &eraserect, 0);
						SDL_UpdateRects(main_surface, 1, &eraserect);
						draw_char(c, x, y, 1);
						x += FONT_WIDTH;
						i++;
					} else if (event.key.keysym.sym == SDLK_BACKSPACE) {
						SDL_Rect	eraserect = {0, y, FONT_WIDTH, FONT_HEIGHT};
						
						i--;
						x -= FONT_WIDTH;
						eraserect.x = x;
						SDL_FillRect(main_surface, &eraserect, 0);
						SDL_UpdateRects(main_surface, 1, &eraserect);
						str[i] = 0;
					}
				}
			}
		}
		SDL_Delay(10);
		// lampeggia il cursore ogni 0.2 secondi
		if ((SDL_GetTicks() - lastblink) > 200) {
			SDL_Rect	blink_rect = {x, y, FONT_WIDTH/5, FONT_HEIGHT};
			
			if (blinker)
				SDL_FillRect(main_surface, &blink_rect, 1);
			else
				SDL_FillRect(main_surface, &blink_rect, 0);
			SDL_UpdateRects(main_surface, 1, &blink_rect);
			blinker = !blinker;
			lastblink = SDL_GetTicks();
		}
	}
}

static int ask_int(char *prompt, int x, int y)
{
	char	str[100] = "";

	ask_string(prompt, str, 99, x, y);
	return atoi(str);
}

void parse_args(int argc, char *argv[])
{
	int	i;

	for (i = 1; i < argc; i++) {
		if (!strcmp(argv[i], "-f"))
			sdl_flags |= SDL_FULLSCREEN;
		else if (!strcmp(argv[i], "-h") || !strcmp(argv[i], "--help"))
			show_help();
		else if (!strcmp(argv[i], "-s")) {
			int		j = 0;
			char	*str, first_digit[10] = "";
			
			i++;
			str = argv[i];
			while (str[j] != 'x')
				j++;
			strncpy(first_digit, str, j);
			width = atoi(first_digit);
			height = atoi(str+j+1);
		} else if (!strcmp(argv[i], "-p")) {
			if (argc > i+1) {
				i++;
				whichPalette = atoi(argv[i]);
				printf("Palette: %d\n", whichPalette);
			} else
				show_help();
		} else if (!strcmp(argv[i], "-mandelbrot")) {
			set = MANDELBROT_SET;
		} else if (!strcmp(argv[i], "-julia")) {
			set = JULIA_SET;
		} else if (!strcmp(argv[i], "-bt")) {
			mandel_algo = BOUNDARY_TRACE;
		} else if (!strcmp(argv[i], "-i")) {
			if (argc > i+1) {
				i++;
				gMaxIterations = atoi(argv[i]);
				printf("Iterazioni massime: %d\n", gMaxIterations);
			} else
				show_help();
		} else {
			printf("Opzione \"%s\" sconosciuta, la ignoro\n", argv[i]);
		}
	}
}

void show_help(void)
{
	printf(	"MeTA~Fract - Giugno 2002 Tito Dal Canton, MeTA~LaBS <tito@puranicla.it>\n\n"
			"Sintassi:\n"
			"         metafract [<opzioni>]\n\n"
			"Opzioni:\n"
			"         -mandelbrot : Disegna l'insieme di Mandelbrot (default)\n"
			"         -approx     : Disegna l'insieme di Mandelbrot con un algoritmo diverso\n"
			"         -julia      : Disegna l'insieme di Julia\n"
			"         -f          : Mostra il frattale a pieno schermo\n"
			"         -p <n>      : Seleziona la sfumatura di colori con cui disegnare i frattali\n"
			"         -i <n>      : Imposta il numero massimo di iterazioni a <n>\n"
			"         -s <x>x<y>  : Imposta le dimensioni del frattale a <x>x<y>, p.es. 640x480\n\n");
	exit(0);
}
